//
//  ARGCamera1.swift
//  ARGearSample
//
//  Created by Jaecheol Kim on 2019/11/11.
//  Copyright © 2019 Seerslab. All rights reserved.
//

import Foundation
import AVFoundation
import CoreImage

class ARGCamera:NSObject, ARGCameraProtocol {
    
    private let captureSession = AVCaptureSession()
    private let dataOutputQueue = DispatchQueue.global(qos: .userInteractive) //DispatchQueue(label: "dataOutputQueue")
    private var dataOutputSynchronizer: AVCaptureDataOutputSynchronizer!
    
    private let videoDataOutput = AVCaptureVideoDataOutput()
    private let metadataOutput = AVCaptureMetadataOutput()
    
    var cameraDevice: AVCaptureDevice?
    var cameraSession: AVCaptureSession?
    var cameraConnection: AVCaptureConnection?
    
    var dataBufferHandler: DataBufferHandler?
    var sampleBufferHandler: SampleBufferHandler?
    
    init(camera: CameraDeviceWithPosition) {
        super.init()
        initCamera(with: camera)
    }
    
    func initCamera(with camera: CameraDeviceWithPosition) {
        
        captureSession.beginConfiguration()
        captureSession.sessionPreset = AVCaptureSession.Preset.iFrame1280x720
        
        setupCameraDevice(with:camera)
        
        // setup outputs
        do {
            // video output
            videoDataOutput.videoSettings = [kCVPixelBufferPixelFormatTypeKey as String: Int(kCVPixelFormatType_32BGRA)]
            videoDataOutput.alwaysDiscardsLateVideoFrames = true
            videoDataOutput.setSampleBufferDelegate(self, queue: dataOutputQueue)
            guard captureSession.canAddOutput(videoDataOutput) else { fatalError() }
            captureSession.addOutput(videoDataOutput)
            cameraConnection = videoDataOutput.connection(with: .video)

            // metadata output
            guard captureSession.canAddOutput(metadataOutput) else { fatalError() }
            captureSession.addOutput(metadataOutput)
            if metadataOutput.availableMetadataObjectTypes.contains(.face) {
                metadataOutput.metadataObjectTypes = [.face]
            }


            // synchronize outputs
            dataOutputSynchronizer = AVCaptureDataOutputSynchronizer(dataOutputs: [videoDataOutput, metadataOutput])
            dataOutputSynchronizer.setDelegate(self, queue: dataOutputQueue)
        }

        setupCameraConnections(with:camera)

        captureSession.commitConfiguration()
        
    }
    
    func startCamera() {
        getVideoPermission(permissionHandler: { granted in
          guard granted else {
            fatalError("Permission not granted to use camera.")
          }
            print("\(self.classForCoder)/" + #function)
            if self.captureSession.isRunning {
                print("already running")
                return
            }
            self.captureSession.startRunning()
        })
    }
    
    func stopCamera() {
        print("\(self.classForCoder)/" + #function)
        if !captureSession.isRunning {
            print("already stopped")
            return
        }
        captureSession.stopRunning()
    }
    
    func changeCamera(with cameraPosition: CameraDeviceWithPosition) {
        let wasRunning = captureSession.isRunning
        if wasRunning {
            captureSession.stopRunning()
        }
        captureSession.beginConfiguration()

        setupCameraDevice(with: cameraPosition)
        setupCameraConnections(with: cameraPosition)

        captureSession.commitConfiguration()

        if wasRunning {
            captureSession.startRunning()
        }
    }
    
    private func getVideoPermission(permissionHandler: @escaping (Bool) -> ()) {
        switch AVCaptureDevice.authorizationStatus(for: .video) {
        case .authorized:
          permissionHandler(true)
        case .notDetermined:
          AVCaptureDevice.requestAccess(for: .video, completionHandler: permissionHandler)
        default:
          permissionHandler(false)
        }
    }

    private func setupCameraDevice(with cameraPosition: CameraDeviceWithPosition) {

        cameraDevice = cameraPosition.captureDevice()
        print("selected video device: \(String(describing: cameraDevice))")

        captureSession.inputs.forEach { (captureInput) in
            captureSession.removeInput(captureInput)
        }
        
        let videoDeviceInput = try! AVCaptureDeviceInput(device: cameraDevice!)
        guard captureSession.canAddInput(videoDeviceInput) else { fatalError() }
        captureSession.addInput(videoDeviceInput)
    }
    
    private func setupCameraConnections(with cameraPosition: CameraDeviceWithPosition) {
        cameraConnection = videoDataOutput.connection(with: .video)!
        switch cameraPosition {
        case .front:
            cameraConnection!.isVideoMirrored = false
        default:
            break
        }
    }
}


// MARK: - Camera delegate
extension ARGCamera: AVCaptureVideoDataOutputSampleBufferDelegate {
    func captureOutput(_ output: AVCaptureOutput, didDrop sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {

    }

    // synchronizer
    func captureOutput(_ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        if let sampleBufferHandler = sampleBufferHandler, connection == cameraConnection
        {
            sampleBufferHandler(output, sampleBuffer, connection)
        }
    }
}

extension ARGCamera: AVCaptureDataOutputSynchronizerDelegate {

    func dataOutputSynchronizer(_ synchronizer: AVCaptureDataOutputSynchronizer, didOutput synchronizedDataCollection: AVCaptureSynchronizedDataCollection) {

        guard let syncedVideoData = synchronizedDataCollection.synchronizedData(for: videoDataOutput) as? AVCaptureSynchronizedSampleBufferData else { return }
        guard !syncedVideoData.sampleBufferWasDropped else {
            print("dropped video:\(syncedVideoData)")
            return
        }

        let videoSampleBuffer = syncedVideoData.sampleBuffer

        let syncedMetaData = synchronizedDataCollection.synchronizedData(for: metadataOutput) as? AVCaptureSynchronizedMetadataObjectData

        dataBufferHandler?(videoSampleBuffer, syncedMetaData?.metadataObjects as NSArray?)
    }
}

